__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Optional, Tuple, Type
from urllib.parse import urlparse


class ReceivedCode(BaseException):
    def __init__(self, code: str) -> None:
        logging.info(code)
        self.code = code
        super().__init__()


class OAuthParameterReceiver(BaseHTTPRequestHandler):

    parameter: Optional[str] = None

    def do_GET(self):  # pylint: disable=invalid-name
        request_query = urlparse(self.path).query
        if self.parameter not in request_query:
            # probably trying to GET /favicon.ico
            return

        key_value: Tuple[str, str] = tuple(request_query.split('=', 1))
        try:
            params = dict([key_value])
            self.write_response(200, f'{self.parameter} received', request_query, 'OK, thanks!')
            raise ReceivedCode(params[self.parameter])
        except (KeyError, ValueError):
            logging.warning("could not parse %s from request path '%s'", self.parameter, self.path)
            return

    def write_response(self, status: int, title: str, request_query: str, message: str):
        self.send_response(status)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(
            bytes(f"""
    <html>
      <head><title>{title}</title></head>
      <body><p>{request_query}</p><p>{message}</p></body>
    </html>
    """,
                  'utf-8'))


class OAuthCodeReceiver(OAuthParameterReceiver):
    """
    When registering an app with the Mixcloud API to receive a token, the API
    requires a redirect URL for a callback:
    `https://www.mixcloud.com/oauth/authorize?client_id=YOUR_CLIENT_ID
                                             &redirect_uri=YOUR_REDIRECT_URI`
     This URL is then called in the form
     `YOUR_REDIRECT_URI?code=OAUTH_CODE`
     This class starts a local HTTP server to catch the callback and extract
     the `code` parameter as the OAuth code.
    """
    parameter = 'code'


class OAuthTokenReceiver(BaseHTTPRequestHandler):
    """
    When registering an app with the Mixcloud API to receive a token, the API
    requires a redirect URL for a callback:
    `https://www.mixcloud.com/oauth/access_token?client_id=YOUR_CLIENT_ID
                                                &client_secret=YOUR_CLIENT_SECRET
                                                &code=OAUTH_CODE
                                                &redirect_uri=YOUR_REDIRECT_URI`
     This URL is then called in the form
     `YOUR_REDIRECT_URI?access_token=ACCESS_TOKEN`
     This class starts a local HTTP server to catch the callback and extract
     the `access_token` parameter as the OAuth code.
    """
    parameter = 'access_token'


def run_local_server(server_port: int, capture_class: Type, create_token) -> None:

    host_name = 'localhost'
    web_server = HTTPServer((host_name, server_port), capture_class)
    logging.info('Server started http://%s:%s', host_name, server_port)

    try:
        web_server.serve_forever()
    except KeyboardInterrupt:
        logging.warning('Interrupted manually')
    except ReceivedCode as received:
        logging.info('Received code %s', received.code)
        create_token.set_oauth_code(received.code)
    except KeyError as error:
        logging.error('mismatched response (%s)', error)
    finally:
        web_server.server_close()

    logging.info('Server stopped.')



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import re
from pathlib import Path
from typing import Tuple, List, Optional

from tinytag import TinyTag


class MixPath:

    def __init__(
            self, basedir: Path, patterns: Tuple[str, ...], exclude_regex: str = r'mix_?\d?.mp3'
    ) -> None:
        self.basedir = basedir
        self.audio_files = sorted([
            f for p in patterns for f in self.basedir.glob(p)
            if not re.search(exclude_regex, str(f))
        ])

    def from_files(self, files: Tuple[Path, ...]):
        return MixPath(self.basedir, tuple(file.name for file in files))

    def file(self, filename: Path) -> Path:
        return self.basedir / filename

    def scan(self) -> Tuple[List[str], int]:
        length = sum(TinyTag.get(str(audio_file)).duration for audio_file in self.audio_files)
        return [f.name for f in self.audio_files], int(length)

    @property
    def tags(self) -> List[str]:
        if (self.basedir / 'tags.txt').exists():
            with (self.basedir / 'tags.txt').open() as file:
                return list(line for line in (line.strip() for line in file) if line)
        return []

    @property
    def title(self) -> str:
        return re.sub(r'^\d+ - ', '', self.basedir.resolve().name)

    @property
    def description(self) -> str:
        if (self.basedir / 'description.txt').exists():
            with (self.basedir / 'description.txt').open() as file:
                return file.read().strip()
        return ''

    @property
    def picture(self) -> Optional[Path]:
        """Currently just returns the first JPEG or PNG. Raises StopIteration if there is none."""
        return next(self.basedir.glob('*.*p*g'))



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import os.path
import re
from math import log10
from pathlib import Path
from shutil import copy2, SameFileError
from typing import List, Optional

from media_tools.util import AudaciousTools


class PlaylistCopier:
    def __init__(self, audacious: AudaciousTools, playlist_id: str) -> None:
        self.audacious = audacious
        self.playlist_id = playlist_id

    def copy_playlist(self, number: int, target: Path, renumber: bool = False) -> None:
        if not target.exists():
            target.mkdir()
        if not target.is_dir():
            raise ValueError(f'{target} is not a directory')
        playlist_id = self.playlist_id or self.audacious.get_currently_playing_playlist_id()
        copy_files(self.audacious.get_existing_files(number, playlist_id), target, renumber)

    def move_files_to_original_places(self, music_dir: Path = Path.home() / 'Music') -> None:
        playlist_id = self.playlist_id or self.audacious.get_currently_playing_playlist_id()
        for file in self.audacious.files_in_playlist(playlist_id):
            move_file(Path(file), music_dir)


def copy_files(filenames_to_copy: List[str], target_dir: Path, renumber: bool) -> None:
    for i, file in enumerate(filenames_to_copy):
        filename = os.path.basename(file)
        target_filename = renumber_file(filename, i + 1, len(filenames_to_copy)) if renumber \
            else filename
        logging.info('%s/%s: %s', i + 1, len(filenames_to_copy), target_filename)
        copy_file(Path(file), target_dir / target_filename)


def find_file_by_name(name: str, path: Path) -> Optional[Path]:
    all_files = sorted(path.rglob(name))
    return all_files[0] if all_files else None


def copy_file(file: Path, target: Path) -> None:
    if not file.exists():
        logging.warning('%s does not exist, skipping', file)
        return
    try:
        copy2(file, target)
    except SameFileError as error:
        logging.warning(str(error))


def renumber_file(filename: str, number: int, total: int) -> str:
    width = max(int(log10(total)) + 1, 2)
    return f'{number:0{width}d} - {strip_leading_numbers(filename)}'


def move_file(file: Path, music_dir: Path):
    if file.is_file():
        return
    target_dir = file.parent
    original_file = find_file_by_name(file.name, music_dir)
    if original_file is None:
        return

    files_to_move = [
        f.name for f in original_file.parent.iterdir() if f.is_file()
    ]
    logging.info('TO MOVE: %s %s %s', original_file, target_dir, files_to_move)

    target_dir.mkdir(exist_ok=True)
    for move in files_to_move:
        (original_file.parent / move).rename(target_dir / move)
        logging.warning(
            '        MOVING %s -> %s', original_file.parent / move, target_dir / move
        )
    original_file.parent.rmdir()


def strip_leading_numbers(filename: str) -> str:
    return re.sub(r'^\d+\s*[-.]?\d*\s*[-.]?\s*', '', filename)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import urllib.error
import urllib.request
from io import BytesIO

from lxml import etree

from media_tools.util.buying.retry_http import retry_http
from media_tools.util.buying.track import Track


class Distributor:

    def search_url(self, track: Track) -> str:
        raise NotImplementedError()

    def is_present(self, track: Track) -> bool:
        raise NotImplementedError()


class Beatport(Distributor):

    def search_url(self, track: Track) -> str:
        return f'http://www.beatport.com/search?q="{track.artist}"+"{track.title}"'

    @retry_http()
    def is_present(self, track: Track) -> bool:
        with urllib.request.urlopen(self.search_url(track)) as response:
            html = etree.parse(BytesIO(response.read()), etree.HTMLParser())
            if html.xpath('//ul[@class ="bucket-items"]'):
                return True
            return bool(html.xpath('//h2[contains(., "Tracks")]'))


class Amazon(Distributor):

    def search_url(self, track: Track) -> str:
        return 'https://smile.amazon.de/s/ref=nb_sb_noss?url=search-alias%3Daps&' \
               f'field-keywords="{track.artist}"+"{track.title}"'

    @retry_http()
    def is_present(self, track: Track) -> bool:
        with urllib.request.urlopen(self.search_url(track)) as response:
            html = etree.parse(BytesIO(response.read()), etree.HTMLParser())
            return bool(html.xpath("//ul[@id='s-results-list-atf']"))



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from pathlib import Path
from typing import Optional

import audioread
import eyed3
import soundfile
from audioread.exceptions import NoBackendError
from mutagen import File, MutagenError


class LengthReader:  # pylint: disable=too-few-public-methods
    def __init__(self, track: Path):
        self.track = track

    def get_duration(self) -> Optional[float]:
        raise NotImplementedError


class MutagenLengthReader(LengthReader):
    def get_duration(self) -> Optional[float]:
        try:
            audio = File(self.track)
            if audio and audio.info:
                return audio.info.length
        except MutagenError:
            pass
        return None


class Eyed3LengthReader(LengthReader):
    def get_duration(self) -> Optional[float]:
        log_level = logging.getLogger().getEffectiveLevel()
        try:
            logging.getLogger().setLevel(logging.ERROR)
            mp3file = eyed3.load(self.track)
            if mp3file is not None and mp3file.info is not None:
                duration = mp3file.info.time_secs
                return duration
        except IOError:
            pass
        finally:
            logging.getLogger().setLevel(log_level)
        return None


class SoundfileLengthReader(LengthReader):
    def get_duration(self) -> Optional[float]:
        try:
            with soundfile.SoundFile(self.track) as mp3file:
                return mp3file.frames / mp3file.samplerate
        except (RuntimeError, ZeroDivisionError):
            pass
        return None


class AudioreadLengthReader(LengthReader):
    def get_duration(self) -> Optional[float]:
        try:
            with audioread.audio_open(self.track) as mp3file:
                return mp3file.duration
        except (NoBackendError, EOFError, UnicodeEncodeError):
            pass
        return None



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from argparse import Namespace
from datetime import date, datetime
from typing import Optional, Sequence

from pylast import LastFMNetwork, User


class LastFMUser:

    @staticmethod
    def from_options(options: Namespace) -> 'LastFMUser':
        network = LastFMNetwork(
            api_key=options.lastfm_api_key, api_secret=options.lastfm_api_secret
        )
        return LastFMUser(network.get_user(options.lastfm_user))

    def __init__(self, user: User) -> None:
        self.user = user

    def start_year(self) -> int:
        return date.fromtimestamp(self.user.get_unixtime_registered()).year

    def get_year_list(self, year: Optional[int]) -> Sequence[int]:
        return [year] if year or year is None else range(self.start_year(), datetime.now().year)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import sys
import threading
import webbrowser
from argparse import Namespace
from functools import partial
from time import sleep
from typing import Dict, Optional, Type
from urllib.parse import urlencode

from media_tools.util.mixcloud.constants import ACCESS_TOKEN_FILE
from media_tools.util.mixcloud.oauth_parameter_receiver import (
    OAuthCodeReceiver, OAuthTokenReceiver, run_local_server
)

MIXCLOUD_URL = 'https://www.mixcloud.com'
REGISTER_APP_URL = f'{MIXCLOUD_URL}/developers/create'
REQUEST_OAUTH_CODE_URL = f'{MIXCLOUD_URL}/oauth/authorize'
REQUEST_TOKEN_URL = f'{MIXCLOUD_URL}/oauth/access_token'


class CreateAPIToken:

    help_text: Optional[str] = None

    def __repr__(self):
        return str(vars(self))

    def pre_run(self):
        print(self.help_text)

    def run(self):
        raise NotImplementedError()

    def post_run(self):
        raise NotImplementedError()

    @classmethod
    def create(cls, opts: Namespace):
        if opts.create_app:
            return Step1()
        if opts.create_code:
            return Step2(opts.client_id)
        if opts.create_token:
            return Step3(opts.client_id, opts.client_secret, opts.oauth_code)
        return Help()


class Help:
    help_text = """
Creating an authentication token for the Mixcloud API takes three steps.

Step 1: Create a Mixcloud app
Before you start this step, ensure that you are logged in to Mixcloud on your
system browser, because you need to fill in the form that is going to open in
your browser.
Then enter:
    mixcloud create-auth --create-app
A window will open in your browser. Follow the instructions printed in your
terminal - click on the app name in the browser and copy the client ID for steps
2 and 3 and the client secret for step 3.

Step 2: Create OAuth token
Use the client ID and secret from step 1 to run
    mixcloud create-auth --create-code --client-id CLIENT_ID
Again a browser window will open with a page displaying an "Authorize" button.
Click it and you will be taken to a page displaying a so-called OAuth code.
Copy the code to use it in step 3.

Step 3: Create the Mixcloud API token
Use the code from step 2 to run
    mixcloud create-auth --create-token --client-id CLIENT_ID \\
                                        --client-secret CLIENT_SECRET \\
                                        --oauth-code OAUTH_CODE
This will open yet another browser window showing you a token (in JSON format).
Copy only the string inside the quotes to use as token - either directly (here
called <TOKEN>) or written (without any extra spaces or newlines) to a file
(here called <TOKEN_FILE>).

Afterwards you can use the generated token to upload mixes to Mixcloud with:
    mixcloud upload --auth-token-string <TOKEN> ...
    mixcloud upload --auth-token-file <TOKEN_FILE> ...
If you store this access token as ".mixcloud_access_token" in either the current
directory, your user's $HOME directory or under $HOME/.config/media-tools, that
token will be used without specifying --auth-token-string or --auth-token-file.
        """

    def run(self):
        print(self.help_text)
        sys.exit(1)


class Step1(CreateAPIToken):

    help_text = """
Please fill in the "Create new app" form opening in your browser.

Please note that you have to be logged in to Mixcloud for the form to work. If
you are not logged in, the "Create" button will be deactivated.

After successfully registering your app it will appear in the "Registered
Applications" section at the top of the page.

Clicking on the app link leads to a page on which you can find the client ID
and the client secret. These will be necessary for the next step.
"""

    def run(self):
        self.pre_run()
        webbrowser.open_new_tab(REGISTER_APP_URL)
        self.post_run()

    def post_run(self):
        pass


class ReceivesResponseViaRedirect(CreateAPIToken):

    request_url: Optional[str] = None
    capture_class: Optional[Type] = None

    def run(self):
        self.pre_run()
        http_server_thread = threading.Thread(
            target=partial(
                run_local_server, capture_class=self.capture_class,
                server_port=self.get_server_port(), create_token=self
            )
        )
        http_server_thread.start()
        sleep(0.1)
        uri = self.get_url()
        webbrowser.open_new_tab(uri)
        http_server_thread.join()
        self.post_run()

    def post_run(self):
        raise NotImplementedError()

    @staticmethod
    def get_server_port() -> int:
        return 8888

    def request_params(self):
        raise NotImplementedError()

    def get_url(self) -> str:
        return f'{self.request_url}?{urlencode(self.request_params())}'

    def local_server(self):
        return f'http://localhost:{self.get_server_port()}'


class Step2(ReceivesResponseViaRedirect):

    request_url = REQUEST_OAUTH_CODE_URL
    capture_class = OAuthCodeReceiver
    help_text = """
Please click the "Authorize" button on the page opening in your browser.
"""

    def __init__(self, client_id: str) -> None:
        self.client_id = client_id
        self.oauth_code: Optional[str] = None

    def set_oauth_code(self, code: str) -> None:
        self.oauth_code = code

    def post_run(self):
        print(f"""
Your OAuth code is {self.oauth_code}. Now start
$ mixcloud create-auth --create-token \
    --client-id {self.client_id} --client-secret <YOUR CLIENT SECRET> \
    --oauth-code {self.oauth_code}
        """)

    def request_params(self) -> Dict[str, str]:
        return {
            'client_id': self.client_id,
            'redirect_uri': self.local_server()
        }


class Step3(ReceivesResponseViaRedirect):

    request_url = REQUEST_TOKEN_URL
    capture_class = OAuthTokenReceiver

    def __init__(self, client_id: str, client_secret: str, oauth_code: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self.oauth_code = oauth_code
        self.token: Optional[str] = None

    def post_run(self):
        print(f"""
Your Mixcloud authorization token is {self.token}. It has been written to the
file ./{ACCESS_TOKEN_FILE}. You can leave it there, or move it to
~/{ACCESS_TOKEN_FILE} or ~/.config/{ACCESS_TOKEN_FILE}.
        """)

    def request_params(self):
        return {
            'client_id': self.client_id,
            'redirect_uri': self.local_server(),
            'client_secret': self.client_secret,
            'code': self.oauth_code
        }



# pylint: disable=unsupported-membership-test,not-an-iterable
__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import webbrowser
from typing import Dict, Optional, Tuple
from urllib.parse import quote_plus

from media_tools.util.buying.trackdb import TrackDB


class Track:

    distributors: Tuple = None
    track_db: TrackDB = None

    @classmethod
    def setup(cls, distributors: Tuple, track_db: Optional[TrackDB]) -> None:
        cls.distributors = distributors
        cls.track_db = track_db

    @classmethod
    def from_dict(cls, data: Dict) -> 'Track':
        try:
            return Track(
                data['artist'].get('name', data['artist'].get('#text')),
                data['name'], int(data.get('playcount', 1))
            )
        except KeyError:
            print("Couldn't create Track from", data)
            raise

    def __init__(self, artist: str, title: str, play_count: int = 1) -> None:
        if self.distributors is None:
            raise ValueError('Call Track.setup() before trying to create a Track!')
        self._artist: str = artist
        self._title: str = title
        self.play_count: int = play_count
        self._buy_url: Optional[str] = None

    def __repr__(self) -> str:
        return f'{self._artist} / {self._title} ({self.play_count} plays)'

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return hash(self._artist + self._title)

    @property
    def title(self) -> str:
        return quote_plus(self._title)

    @property
    def artist(self) -> str:
        return quote_plus(self._artist)

    def buy_url(self) -> str:
        if self._buy_url is not None:
            return self._buy_url

        self._buy_url = ''
        for distributor in self.distributors:
            if distributor.is_present(self):
                self._buy_url = distributor.search_url(self)
                break
        return self._buy_url

    def buy(self) -> None:
        if self.buy_url() and Track.track_db is not None and self not in Track.track_db:
            webbrowser.open(self.buy_url())
            Track.track_db.add(self)



import re
from pathlib import Path
from typing import List

from .base_filename_cleaner import BaseFilenameCleaner
from .command import Command, Commands, Move
from .util import find_files


class JunkRemover(BaseFilenameCleaner):

    JUNK_TO_REMOVE = {
        ' $': '',     # space(s) at beginning and before ".mp3"
        '-$': '',     # "-.mp3", " - .mp3"
        '^-': '',     # dash at beginning
        '^ ': '',     # space at beginning
        r'^\.': '',   # dot at beginning
        '–': '-',     # em dash
        '--': '-',    # --
        '- -': '-',   # stray double dashes
        '_': ' ',     # underscores
        '-,': '-',    # dash before comma (BECAUSE THAT HAPPENS AND IT FUCKS EVERYTHING UP JFC)
        r' -(\S)': r' - \1',  # immediately leading dash with space before
        r'(\S)- ': r'\1 - ',  # immediately trailing dash with space after
        '  ': ' ',    # double spaces
        ' ,': ',',    # space before comma
        r'\[\]': '',  # []
        r'\(\)': '',  # ()
        r'-[\da-f]{8}$': '',  # trailing hex string
    }

    def fix_commands(self) -> Commands:
        def has_junk(filename: Path) -> bool:
            return self.is_music_file(filename) and \
                any(re.search(s, self.filename_base(filename)) for s in self.JUNK_TO_REMOVE)

        mismatches = sorted(find_files(self._base_directory, has_junk))
        return [self.fix_command_for_file(mismatch) for mismatch in mismatches]

    def fix_command_for_file(self, mismatch: Path) -> Command:
        root = mismatch.parent
        fixed = self.filename_base(mismatch)
        extension = mismatch.suffix
        matches: List[str] = []
        changed = True
        while changed:
            changed, fixed = self._remove_some_junk(fixed, matches)
        return Move(mismatch, root.joinpath(fixed + extension), matches)

    def _remove_some_junk(self, fixed: str, matches: List[str]):
        changed = False
        for search, replace in self.JUNK_TO_REMOVE.items():
            new_fixed = re.sub(search, replace, fixed)
            if new_fixed != fixed:
                changed = True
                matches.append(search)
            fixed = new_fixed
        return changed, fixed



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import os
import sys
from argparse import ArgumentParser, Namespace
from typing import List

from media_tools.util.scrobbles import LastFMUser, Scrobbles
from media_tools.util.logging import setup_logging

MAX_RETRY_WAIT = 1024


def parse(args: List[str]) -> Namespace:

    def env_or_required(key):
        return {'default': os.environ[key]} if key in os.environ else {'required': True}

    parser = ArgumentParser(
        description="Creates a pickle or JSON file from a last.fm user's scrobbles"
    )
    parser.add_argument(
        '-l', '--limit', type=int, default=None,
        help='Max. number of entries to request'
    )
    parser.add_argument(
        '-y', '--year', type=int, default=None,
        help='Year the data is requested for (empty for entire period)'
    )
    parser.add_argument(
        '-o', '--output', type=str, default=None,
        help='output file name'
    )
    parser.add_argument(
        '--lastfm-user', type=str,
        help='Last.FM user', **env_or_required('PYLAST_USERNAME')
    )
    parser.add_argument(
        '--lastfm-api-key', type=str,
        help='Last.FM API key', **env_or_required('PYLAST_API_KEY')
    )
    parser.add_argument(
        '--lastfm-api-secret', type=str,
        help='Last.FM API secret', **env_or_required('PYLAST_API_SECRET')
    )
    parser.add_argument(
        '-m', '--max-retry-wait', type=int, default=MAX_RETRY_WAIT,
        help='Max. seconds to wait between requests to avoid rate limiting'
    )
    return parser.parse_args(args)


def main(args: List[str]) -> None:
    options = parse(args)
    setup_logging(options)

    user = LastFMUser.from_options(options)

    for year in user.get_year_list(options.year):
        scrobble_data = Scrobbles(user, year, options.limit, options.max_retry_wait)
        scrobble_data.get_and_dump(options.output)


def entry() -> None:
    main(sys.argv[1:])


if __name__ == '__main__':
    entry()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from pathlib import Path
from typing import List

from media_tools.util.base_filename_cleaner import BaseFilenameCleaner
from media_tools.util.command import Commands, Move


class ExtensionLowercaser(BaseFilenameCleaner):

    def fix_commands(self) -> Commands:
        return [
            Move(to_lower, self.lowercased_ext(to_lower), [])
            for to_lower in self.find_uppercase_extensions()
        ]

    def find_uppercase_extensions(self) -> List[Path]:
        return [
            with_uppercase
            for with_uppercase in self.get_music_files()
            if with_uppercase.suffix.lower() != with_uppercase.suffix
        ]

    @staticmethod
    def lowercased_ext(file: Path) -> Path:
        return file.parent / Path(file.stem).with_suffix(file.suffix.lower())



import re
from pathlib import Path
from typing import Optional

from media_tools.util.base_filename_cleaner import BaseFilenameCleaner
from media_tools.util.command import Command, Commands, Move, Nothing
from media_tools.util.media_extensions import MEDIA_EXTENSIONS
from media_tools.util.util import find_files, log_utf8_safe


class NumberingFixer(BaseFilenameCleaner):

    # Note: The order of patterns matters. Patterns with more specific matches must come first.
    PATTERNS_TO_FIX = [
        r'(\d\d)\2\s+-\s+([^/]+)',       # 0101 - blah (note: regex is prefixed by group for path)
        r'\s*(\d{1,3}-\d)\s+([^-][^/]+)',    # 01-1 blah
        r'\s*(\d-\d{1,2})\s+([^-][^/]+)',    # 1-01 blah
        r'\s*(\d-\d{1,2})\.\s+([^-][^/]+)',  # 1-01. blah
        r'\s*(\d-\d{1,2})\.([^-][^/]+)',     # 1-01.blah
        r'\s*(\d\d-\d\d)\s+([^-][^/]+)',     # 01-01 blah
        r'\s*(\d{1,3}) – ([^/]+)',           # 01 – blah (en dash)
        r'\s*(\d{1,3})–([^/]+)',             # 01–blah (en dash)
        r'\s*(\d{1,3}) — ([^/]+)',           # 01 — blah (em dash)
        r'\s*(\d{1,3})—([^/]+)',             # 01—blah (em dash)
        r'\s*(\d{1,4})\s+([^/]+)',           # 01 blah
        r'\s*(\d{1,3})\.\s+([^/]+)',         # 01. blah
        r'\s*(\d{1,3})\.([^/]+)',            # 01.blah
        r'\s*(\d{1,3})--([^/]+)',            # 01--blah
        r'\s*(\d{1,3})-\s*(\D[^/]+)',        # 01-blah or 01- blah
        r'\s*-(\d{1,3})-([^/]+)',            # -01-blah
        r'\s*-\s*(\d{1,3})\.\s*([^/]+)',     # - 01. blah
        r'\s*(\d{1,3})_([^/]+)',             # 01_blah
        r'\s*\[(\d{1,3})\]\s+([^/]+)',       # [01] blah, [01]  blah, ...
        r'\s*\[(\d{1,3})\]([^/]+)',          # [01]blah
        r'\s*(\d{1,3})\]([^/]+)',            # 01]blah
        r'\s*\((\d{1,3})\)\s*([^/]+)',       # (01)blah
        r'\s*(\d{1,4})(\D[^/]*)',            # 01blah
        r'\s*([a-z]\d{1,2})\s+([^/]+)',      # a1 blah
        r'\s*([a-z]\d)-([^/]+)',             # a1-blah
        r'\s*([a-z]\d)\.([^/]+)',            # a1.blah
        r'\s*\[([a-z]\d)\]([^/]+)',          # [a1]blah
        r'\s*([a-z]\d)\]([^/]+)',            # a1]blah
        r'\s*\(([a-z]\d)\)([^/]+)',          # (a1)blah
        r'\s*([a-z]\d{1,2})(\D[^/]+)',       # a1blah
    ]
    PATTERNS_TO_FIX_2 = [
        r'\s*(\d) - (\d{1,2})\.\s*([^-][^/]+)',    # 1 - 01. blah, 1 - 01.blah
        r'\s*(\d) - (\d{1,2})\s*-\s*([^-][^/]+)',  # 1 - 01-blah, 1 - 01 - blah
        r'\s*(\d)\.(\d{1,2})\s*-*\s*([^-][^/]+)',  # 1.01 blah, 1.01-blah, 1.01 - blah,
        r'\s*(\d)-(\d{1,2})\.\s*([^-][^/]+)',      # 1-01.blah, 1-01. blah
        r'\s*(\d)-(\d{1,2})\s*-*\s*([^-][^/]+)',   # 1-01-blah, 1-01 - blah
    ]

    def fix_commands(self) -> Commands:
        def has_screwy_numbering(filename: Path) -> bool:
            base = self.filename_base(filename)
            return self.is_music_file(filename) and bool(
                re.search(r'\d+', base) and
                not re.search(r'^\d{1,4} - [^/]+', base, flags=re.IGNORECASE) and
                not re.search(r'^\d{1,4}$', base, flags=re.IGNORECASE) and
                not re.search(r'^\d{1,2}-\d{1,2}$', base, flags=re.IGNORECASE) and
                not re.search(r'^\d{1,2}-\d{1,2} - [^/]+', base, flags=re.IGNORECASE) and
                not re.search(r'^[a-z]\d{1,2} - [^/]+', base, flags=re.IGNORECASE) or
                re.search(r'^(\d\d)\1 - [^/]+', base, flags=re.IGNORECASE) or
                re.search(r'^\d - \d\d.\s*[^/]+', base, flags=re.IGNORECASE) or
                re.search(r'^\d - \d\d\s*-\s*[^/]+', base, flags=re.IGNORECASE) or
                re.search(r'\d{1,4}\s?–', base, flags=re.IGNORECASE) or
                re.search(r'\d{1,4}\s?—', base, flags=re.IGNORECASE)
            )

        mismatches = sorted(find_files(self._base_directory, has_screwy_numbering))
        return [self._fix_numbering_for_file(file) for file in mismatches]

    def _fix_numbering_for_file(self, file: Path) -> Command:
        for extension in MEDIA_EXTENSIONS:
            result = self._may_fix_numbering_for_file_and_extension(str(file), extension)
            if result is not None:
                return result
        log_utf8_safe('-' * 8, file)
        return Nothing()

    def _may_fix_numbering_for_file_and_extension(
            self, filename: str, extension: str
    ) -> Optional[Command]:
        # just a number - do not change
        if re.match(r'(.*)/(\d{1,4})\.' + extension, filename, flags=re.IGNORECASE):
            return Nothing()
        # CD id, track number and track name
        for pattern in self.PATTERNS_TO_FIX_2:
            match = re.search(
                '(.*)/' + pattern + r'\.' + extension, filename, flags=re.IGNORECASE
            )
            if match:
                return Move(
                    Path(filename),
                    Path(match.group(1)) /
                    Path(f'{match.group(2)}-{match.group(3)} - {match.group(4)}.{extension}'),
                    [pattern]
                )
        # track number and track name
        for pattern in self.PATTERNS_TO_FIX:
            match = re.search(
                '(.*)/' + pattern + r'\.' + extension, filename, flags=re.IGNORECASE
            )
            if match:
                print(match.groups(), pattern)
                return Move(
                    Path(filename),
                    Path(f'{match.group(1)}/{match.group(2)} - {match.group(3)}.{extension}'),
                    [pattern]
                )
        return None



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from calendar import timegm
from datetime import datetime
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Union

from pylast import PyLastError

from media_tools.util.dump_data import dump_data
from media_tools.util.lastfm_user import LastFMUser


class Scrobbles:

    def __init__(
            self, user: LastFMUser, year: Optional[int], limit: Optional[int], max_retry_wait: int
    ):
        self.user = user.user
        self.year = year
        self.limit = limit
        self.max_retry_wait = max_retry_wait

    def get_and_dump(self, filename: Optional[str] = None):
        scrobble_data = self.get_data()
        backup_filename = get_output_filename(self.year, filename)
        dump_data(scrobble_data, backup_filename)

    def get_data(
            self, wait_time: int = 0, tracks: Optional[List[str]] = None,
            loved: Optional[List[str]] = None, top_tracks: Optional[List[str]] = None
    ) -> Dict:
        logging.debug(
            'get_data(user=%s, year=%s, limit=%s, wait=%s)',
            self.user.name, self.year, self.limit, wait_time
        )
        year_dates = year_range(self.year)
        try:
            play_count = self.user.get_playcount()
            tracks = self.get_and_wait(tracks, self.user.get_recent_tracks, wait_time, **year_dates)
            loved = self.get_and_wait(loved, self.user.get_loved_tracks, wait_time)
            top_tracks = self.get_and_wait(top_tracks, self.user.get_top_tracks, wait_time)
            top_artists = self.get_and_wait(None, self.user.get_top_artists, wait_time)
            top_albums = self.get_and_wait(None, self.user.get_top_albums, wait_time)
            top_tags = self.user.get_top_tags(limit=self.limit)
            return {
                'time': datetime.now(),
                'play_count': play_count,
                'tracks': tracks,
                'loved': loved,
                'top_tracks': top_tracks,
                'top_artists': top_artists,
                'top_albums': top_albums,
                'top_tags': top_tags,
            }
        # pylast has a feature to deal with rate limiting, but it doesn't seem to work reliably
        except PyLastError as error:
            logging.info('%s: %s', type(error).__name__, str(error))
            next_wait_time = 1 if wait_time == 0 else 2 * wait_time
            if next_wait_time <= self.max_retry_wait:
                logging.warning(
                    '%s, retrying with %ss wait time', type(error).__name__, next_wait_time
                )
                return self.get_data(
                    next_wait_time, tracks=tracks, loved=loved, top_tracks=top_tracks
                )
            logging.error('Rate limiting still active after %ss - giving up', self.max_retry_wait)
            raise

    def get_and_wait(
            self, values: Optional[List], function: Callable, wait_time: int, **kwargs
    ) -> Union[List[Any], Any]:
        if values is None:
            values = function(limit=self.limit, **kwargs)
            logging.info('waiting %ss', str(wait_time))
            sleep(wait_time)
        return values


def get_output_filename(year: Optional[int], filename: Optional[str]) -> str:
    if filename is None:
        year_string = 'all' if year is None else str(year)
        filename = f'backup_lastfm_{year_string}.pickle.bz2'
    return filename


def from_datetime(convert_datetime: datetime) -> int:
    return timegm(convert_datetime.utctimetuple())


def year_range(year: Optional[int]) -> Dict[str, int]:
    start = None if year is None else from_datetime(datetime(year, 1, 1, 0, 0, 0))
    end = None if year is None else from_datetime(datetime(year, 12, 31, 23, 59, 59, 999999))
    return {'time_from': start, 'time_to': end}



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import sys
from argparse import ArgumentError, ArgumentParser, Namespace
from pathlib import Path
from shutil import copy
from typing import Callable, List

from media_tools.util.logging import setup_logging
from media_tools.util.mixcloud import (
    AuthorizationError, create_mix, get_access_token, DEFAULT_CROSSFADE_MS, DEFAULT_MAX_RETRY,
    DEFAULT_AUDIO_FILE_TYPES,
    MixPath
)
from media_tools.util.mixcloud.create_api_token import CreateAPIToken


def exists(check: Callable) -> Callable:
    def inner(arg: str) -> Path:
        path = Path(arg)
        if not path.exists():
            raise ValueError(f'{arg} does not exist')
        if not check(path):
            raise ValueError(f'{arg} is of wrong type')
        return path
    return inner


def parse_commandline(args: List[str]) -> Namespace:
    exit_on_error = {'exit_on_error': False} if sys.version_info >= (3, 9) else {}
    parser = ArgumentParser(
        description='Interface to the Mixcloud API to create mixes',
        **exit_on_error  # type: ignore
    )
    subparsers = parser.add_subparsers(
        title='subcommands', required=True
    )
    parse_upload_args(exit_on_error, subparsers)
    parse_create_auth_args(exit_on_error, subparsers)

    return parser.parse_args(args)


def parse_upload_args(exit_on_error, subparsers):
    parser_upload = subparsers.add_parser(
        'upload', help='Create a mix from audio files and uploads it to Mixcloud',
        **exit_on_error  # type: ignore
    )
    parser_upload.add_argument(
        '-d', '--directory', type=exists(Path.is_dir), required=True,
        help='Directory containing the mix'
    )
    parser_upload.add_argument(
        '-e', '--extensions', nargs='+', default=DEFAULT_AUDIO_FILE_TYPES,
        help='List of extensions considered as audio files for the mix'
    )
    parser_upload.add_argument(
        '-q', '--quiet', action='store_true'
    )
    parser_upload.add_argument(
        '-s', '--strict', action='store_true', help='Fail if any required data are missing'
    )
    parser_upload.add_argument(
        '-c', '--crossfade-ms', type=int, default=DEFAULT_CROSSFADE_MS,
        help='Milliseconds overlap between tracks'
    )
    parser_upload.add_argument(
        '-r', '--max-retry', type=int, default=DEFAULT_MAX_RETRY,
        help='Maximum number of retries for failing uploads'
    )
    group_auth = parser_upload.add_mutually_exclusive_group()
    group_auth.add_argument(
        '-a', '--auth-token-file', type=exists(Path.is_file),
        help='File containing the Mixcloud auth token'
    )
    group_auth.add_argument(
        '-t', '--auth-token-string', type=str, help='Mixcloud auth token as string'
    )
    group_description = parser_upload.add_mutually_exclusive_group()
    group_description.add_argument(
        '--description', type=str, help='Description for the mix as string'
    )
    group_description.add_argument(
        '--description-file', type=exists(Path.is_file),
        help='File containing the description for the mix'
    )
    group_tags = parser_upload.add_mutually_exclusive_group()
    group_tags.add_argument(
        '--tags', nargs='*', type=str, help='Tags for the mix as a list of strings'
    )
    group_tags.add_argument(
        '--tags-file', type=exists(Path.is_file),
        help='File containing the tags for the mix, one per line'
    )
    parser_upload.add_argument(
        '--picture-file', type=exists(Path.is_file), help='Picture file for the mix'
    )
    parser_upload.set_defaults(execute=do_upload)


def parse_create_auth_args(exit_on_error, subparsers):
    parser_create_auth = subparsers.add_parser(
        'create-auth', help='Create Mixcloud authentication and authorization',
        **exit_on_error  # type: ignore
    )
    parser_create_auth.add_argument(
        '--browser', choices=('system', 'selenium'), default='system',
        help='Browser used for app and token registration ("system" for interactive, '
             '"selenium" for automatic)'
    )
    group_step = parser_create_auth.add_mutually_exclusive_group()
    group_step.add_argument(
        '--create-app', action='store_true', help='Create a Mixcloud app (step 1)',
    )
    group_step.add_argument(
        '--create-code', action='store_true', help='Create an OAuth code (step 2)'
    )
    group_step.add_argument(
        '--create-token', action='store_true', help='Create an OAuth token (step 3)'
    )
    parser_create_auth.add_argument(
        '--client-id', type=str, help='Client ID for the app you generated in step 1'
    )
    parser_create_auth.add_argument(
        '--client-secret', type=str, help='Client ID for the app you generated in step 1'
    )
    parser_create_auth.add_argument(
        '--oauth-code', type=str, help='OAuth code you received in step 2'
    )
    parser_create_auth.set_defaults(execute=create_authentication)


def prep_mix_dir(opts: Namespace) -> None:
    description_file = opts.directory / 'description.txt'
    tags_file = opts.directory / 'tags.txt'
    if opts.description_file:
        copy(opts.description_file, description_file)
    if opts.description:
        with description_file.open('w') as file:
            file.write(opts.description)
    if opts.tags_file:
        copy(opts.tags_file, tags_file)
    if opts.tags:
        with tags_file.open('w') as file:
            file.write('\n'.join(opts.tags))
    if opts.picture_file:
        copy(opts.picture_file, opts.directory)
    check_strict(opts, description_file, tags_file)


def check_strict(opts, description_file, tags_file):
    if not description_file.exists():
        warning_or_error(opts, f'description file {description_file} needed')
    if not tags_file.exists():
        warning_or_error(opts, f'tags file {tags_file} needed')
    if not list(opts.directory.glob('*.*p*g')):
        warning_or_error(opts, f'picture file needed in {opts.directory}')


def warning_or_error(opts, message):
    if opts.strict:
        raise ValueError(message)
    logging.warning(message)


def extract_auth_token(opts: Namespace) -> str:
    if opts.auth_token_string is not None:
        return opts.auth_token_string
    return get_access_token(opts.auth_token_file)


def do_upload(opts: Namespace):
    try:
        access_token = extract_auth_token(opts)
        prep_mix_dir(opts)

        mix_path = MixPath(Path(opts.directory), tuple(f'*.{ext}' for ext in opts.extensions))
        mix = create_mix(mix_path, access_token, crossfade_ms=opts.crossfade_ms, strict=opts.strict)
        mix.export()
        mix.upload(max_retry=opts.max_retry)
    except (ValueError, AuthorizationError) as error:
        logging.error(error)
        sys.exit(1)
    except KeyboardInterrupt:
        logging.warning('interrupted by user!')


def create_authentication(opts: Namespace):
    if opts.create_code and not opts.client_id:
        raise ArgumentError(None, message='--create-code needs --client-id')
    if opts.create_token and not (opts.client_id and opts.client_secret and opts.oauth_code):
        raise ArgumentError(
            None, message='--create-code needs --client-id, --client-secret and --oauth-code '
        )
    if opts.browser == 'selenium':
        raise NotImplementedError('selenium engine')
    create_token = CreateAPIToken.create(opts)
    create_token.run()


def main() -> None:
    try:
        opts = parse_commandline(sys.argv[1:])
    except ArgumentError as error:
        setup_logging(Namespace())
        logging.error(error)
        sys.exit(1)

    setup_logging(opts)
    opts.execute(opts)


if __name__ == '__main__':
    main()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import sys
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from time import sleep
from typing import List, Dict, Optional, Tuple

import requests
from requests.adapters import HTTPAdapter, Retry
from pydub import AudioSegment
from tinytag import TinyTag

from media_tools.util.mixcloud.constants import (
    DEFAULT_CROSSFADE_MS, DEFAULT_MAX_RETRY, MP3_KBIT_RATE, MIXCLOUD_API_UPLOAD_URL,
    MIXCLOUD_MAX_TAGS, ALLOWED_TRACKS_PER_ARTIST
)
from media_tools.util.mixcloud.mix_path import MixPath


@dataclass
class RetryOptions:
    max_retry: int = DEFAULT_MAX_RETRY
    backoff_factor: float = 0.5
    status_codes: Tuple[int, ...] = (504, 503, 502, 429, 400, 403)


class Mix:  # pylint: disable=too-many-instance-attributes

    def __init__(
            self, mix_path: MixPath, strict: bool = False, crossfade_ms: int = DEFAULT_CROSSFADE_MS,
            part: Optional[int] = None
    ) -> None:
        self._mix_path = mix_path
        self._strict = strict
        self._crossfade_ms = crossfade_ms
        self._incomplete = False
        self._track_info: List[Dict] = []
        self._title: Optional[str] = None
        self._access_token: Optional[str] = None
        self._part: Optional[int] = part
        if self._strict and not self._mix_path.audio_files:
            raise ValueError(f'no audio files in {self._mix_path.basedir}')
        self._audio = self._import_audio(self._mix_path.audio_files)

    def with_access_token(self, access_token: str) -> 'Mix':
        self._access_token = access_token
        return self

    def with_title(self, title: str) -> 'Mix':
        self._title = title
        return self

    @property
    def valid(self):
        return not self._incomplete

    @property
    def tags(self) -> List[str]:
        tags = self._mix_path.tags
        error = self.check_tag_validity(tags)
        if error:
            self.raise_error(error)
        return tags

    def raise_error(self, error: str) -> None:
        self._incomplete = True
        if self._strict:
            raise ValueError(error)
        logging.warning(error)

    @staticmethod
    def check_tag_validity(tags: List[str]) -> str:
        if len(tags) > MIXCLOUD_MAX_TAGS:
            return f'Max. {MIXCLOUD_MAX_TAGS} tags allowed, found {len(tags)}: {tags}'
        if not tags:
            return 'No tags found'
        return ''

    @property
    def title(self) -> str:
        if self._title:
            return self._title
        title = self._mix_path.title
        return f"Test - don't bother playing ({title})" if self._incomplete else title

    @property
    def description(self) -> str:
        description = self._mix_path.description
        if description:
            return description
        self.raise_error('No description found')
        return ''

    @property
    def picture(self) -> Optional[Path]:
        """Currently just returns the first JPEG or PNG. Room for improvement!"""
        try:
            return self._mix_path.picture
        except StopIteration:
            self.raise_error(f'No picture in {self._mix_path.basedir}')
            return None

    def _import_audio(self, audio_files: List[Path]) -> AudioSegment:
        part_name = '' if self._part is None else f'_{self._part}'
        mix_file = self._mix_path.file(Path(f'mix{part_name}.mp3'))
        self._track_info = [self._get_track_info(audio_file) for audio_file in audio_files]
        self._check_tracks_per_artist_limit()
        if mix_file.exists():
            logging.info('Reading present %s', mix_file)
            return AudioSegment.from_file(mix_file)

        audio = AudioSegment.empty()
        for i, audio_file in enumerate(audio_files):
            logging.info(
                '%s: %s (%s - %s)',
                i + 1, audio_file.name, self._track_info[i]['artist'], self._track_info[i]['title']
            )
            audio = audio.append(
                self._read_track(audio_file),
                crossfade=self._crossfade_ms if len(audio) > self._crossfade_ms else len(audio)
            )

        return audio

    def _check_tracks_per_artist_limit(self) -> None:
        artists: Dict[str, int] = defaultdict(int)
        for track in self._track_info:
            artists[track['artist']] += 1
        if artists and max(artists.values()) > ALLOWED_TRACKS_PER_ARTIST:
            artists_with_too_many_tracks = ', '.join([
                a for a, num in artists.items() if num > ALLOWED_TRACKS_PER_ARTIST
            ])
            self.raise_error(
                f'More than {ALLOWED_TRACKS_PER_ARTIST} tracks by: {artists_with_too_many_tracks}')

    @staticmethod
    def _read_track(audio_file):
        track = AudioSegment.from_file(audio_file)
        if track.sample_width != 2:
            track = track.set_sample_width(2)
        if track.frame_rate != 44100:
            track = track.set_frame_rate(44100)
        return track.normalize()

    def _get_track_info(self, audio_file: Path) -> Dict:
        tags = TinyTag.get(str(audio_file))
        if tags.artist is None or tags.title is None:
            self.raise_error(f'Incomplete tags for {audio_file}')
            return {
                'artist': tags.artist or '???', 'title': tags.title or '???',
                'length': tags.duration
            }
        return {
            'artist': tags.artist, 'title': tags.title, 'length': tags.duration,
            'filename': audio_file.name
        }

    def export(self, name: Path = Path('mix.mp3')) -> None:
        mix_file = self._mix_path.file(name)
        if mix_file.exists():
            logging.info('Mix already exported as %s', mix_file)
            return
        audio_format = name.suffix[1:]
        logging.info(
            'Exporting %s audio to %s with bitrate %s kbps',
            self.mix_length(), mix_file, MP3_KBIT_RATE
        )
        self._audio.export(
            mix_file, format=audio_format, parameters=['-q:a', '0'], bitrate=f'{MP3_KBIT_RATE}k'
        )

    def upload(self, name: Path = Path('mix.mp3'), max_retry: int = DEFAULT_MAX_RETRY) -> None:
        mix_file = self._mix_path.file(name)
        if not mix_file.exists():
            raise FileNotFoundError(mix_file)
        files = {
            'mp3': (str(name), mix_file.open('rb'), 'audio/mpeg'),
        }
        if self.picture:
            picture_type = self.picture.suffix
            files['picture'] = (
                'picture' + picture_type, self.picture.open('rb'),
                f'image/{"png" if picture_type == ".png" else "jpeg"}'
            )

        data = {
            'name': self.title,
            'description': self.description,
            'percentage_music': 100
        }
        self._add_tags(data)
        self._add_track_info(data)
        size_bytes = mix_file.stat().st_size + self.picture.stat().st_size if self.picture else 0
        logging.info(
            'Uploading %s kBytes (%s) as %s',
            f'{size_bytes // 1024:,d}', self.mix_length(), self.title
        )
        self._do_upload(name, files, data, max_retry)

    def mix_length(self) -> str:
        return f'{len(self._audio) // 60000}:{len(self._audio) % 60000 // 1000:02}'

    def _do_upload(self, name: Path, files: Dict, data: Dict, max_retry: int):
        if max_retry < 0:
            logging.error('Out of retries, aborting upload attempt')
            sys.exit(1)

        url = MIXCLOUD_API_UPLOAD_URL + '/?access_token=' + self._access_token
        logging.info(
            'Uploading to %s, max. %s retries',
            url.replace(self._access_token, '<REDACTED>'), max_retry
        )
        try:
            response = self._send_request(
                url, files=files, data=data, retry_options=RetryOptions(max_retry=max_retry)
            )
        except requests.exceptions.ConnectionError as error:
            logging.error(error)
            sys.exit(1)
        else:
            if self.is_rate_limited(response):
                logging.info(self.response_message(response))
                sleep(int(response.json().get('error', {}).get('retry_after', 60)))
                self._do_upload(name, files=files, data=data, max_retry=max_retry - 1)
            elif response.status_code == 200:
                logging.info(self.response_message(response))
                sys.exit(0)
            else:
                logging.error(self.response_message(response))
                sys.exit(1)

    @staticmethod
    def _send_request(
            url: str, files: Dict, data: Dict, retry_options: RetryOptions = RetryOptions()
    ) -> requests.Response:
        sess = requests.Session()
        retries = Retry(
            total=retry_options.max_retry, backoff_factor=retry_options.backoff_factor,
            status_forcelist=retry_options.status_codes
        )
        sess.mount('https://', HTTPAdapter(max_retries=retries))
        return sess.post(url, data=data, files=files)

    @staticmethod
    def is_rate_limited(response: requests.Response) -> bool:
        return (response.status_code == 403 and
                response.json().get('error', {}).get('type') == 'RateLimitException')

    @staticmethod
    def response_message(response: requests.Response) -> str:
        if response.status_code == 200:
            return f"{response.status_code}: {response.json()['result']['message']}"
        return f"{response.status_code}: {response.json()['error']}"

    @staticmethod
    def connection_error_message(error: requests.exceptions.ConnectionError, max_retry: int) -> str:
        message = Mix.error_args_message(error)
        if max_retry > 0:
            return f"{message} Retrying {max_retry} time{'s' if max_retry > 1 else ''}."
        return f'{message} Giving up.'

    @staticmethod
    def error_args_message(error: requests.exceptions.ConnectionError) -> str:
        if not error.args:
            return ''
        if error.args[0].args and len(error.args[0].args) > 1:
            return f'{error.args[0].args[1].args[1]}: {error.args[0].args[0]} '
        return ', '.join(str(arg) for arg in error.args)

    def _add_tags(self, data: Dict) -> None:
        for i, tag in enumerate(self.tags):
            data[f'tags-{i}-tag'] = tag

    def _add_track_info(self, data: Dict) -> None:
        start_time = 0
        for i, track_info in enumerate(self._track_info):
            data[f'sections-{i}-artist'] = track_info['artist']
            data[f'sections-{i}-song'] = track_info['title']
            data[f'sections-{i}-start_time'] = int(start_time)
            start_time += (track_info['length'] - self._crossfade_ms / 1000)
